
/**
* Library for Deferred Shading.
* -----------------------------------
* author: Ryan Schmitt
* date: 2-20-11
*/

#include "DeferredShadingEngine.h"

//Some constants
const char* DeferredShadingEngine::GATHER_VERT_SHADER = "./shaders/deferred_gather.vert";
const char* DeferredShadingEngine::GATHER_FRAG_SHADER = "./shaders/deferred_gather.frag";
const char* DeferredShadingEngine::DIR_TARGET_FRAG_SHADER = "./shaders/deferred_direct_target.frag";
const char* DeferredShadingEngine::DIR_AMB_LIGHT_FRAG_SHADER = "./shaders/dLight.frag";
const char* DeferredShadingEngine::POINT_LIGHT_VERT_SHADER = "./shaders/pLight.vert";
const char* DeferredShadingEngine::POINT_LIGHT_FRAG_SHADER = "./shaders/pLight.frag";

DeferredShadingEngine::DeferredShadingEngine(int w, int h)
: width(w), height(h), gatherShader(NULL), pointLightShader(NULL), dirAmbLightShader(NULL), directTargetShader(NULL)
{
   checkExtensions();
}

DeferredShadingEngine::~DeferredShadingEngine()
{
   if(gatherShader != NULL)
      delete gatherShader;
   if(pointLightShader != NULL)
      delete pointLightShader;
   if(dirAmbLightShader != NULL)
      delete dirAmbLightShader;
   if(directTargetShader != NULL)
      delete directTargetShader;
   glDeleteTextures(NUM_MRTS, textures);
   glDeleteFramebuffersEXT(1, &fbo);
}

void
DeferredShadingEngine::init()
{
   printf("\n***DeferredShadingEngine INITIALIZATION***\n\n");
   //FBO
   glGenFramebuffersEXT(1, &fbo);
   glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
   //Depth
   glActiveTexture(GL_TEXTURE0);
   initDepthTexture();
   //Normals
   initRenderTargetTexture("Normal Map", &textures[NORMALS], GL_COLOR_ATTACHMENT0_EXT);
   //Colors
   initRenderTargetTexture("Color Map", &textures[COLORS], GL_COLOR_ATTACHMENT1_EXT);

   //Set targets to draw buffers
   GLenum attachmentList[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };
   glDrawBuffers(2, attachmentList); //2 because only normals and color atm.

   //Status Check
   GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
   if(status!= GL_FRAMEBUFFER_COMPLETE_EXT) {
      printf("Framebuffer loading error...\n");
      exit(-1);
   }

   glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); //unbind fbo for now

   //Shaders
   printf("\n");
   initShaders();

   printf("\n***DeferredShadingEngine INIT COMPLETED***\n");
}

void
DeferredShadingEngine::initDepthTexture() {
   printf("\tDepth Map activating...");
   glGenTextures(1, &depthtexture);
   glBindTexture(GL_TEXTURE_2D, depthtexture);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT32, width, height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_NONE);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
   glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthtexture, 0);
   printf("DONE\n");
}

void
DeferredShadingEngine::initRenderTargetTexture(const char* name, GLuint* texture, GLenum attachment) {
   printf("\t%s activating...", name);
   glGenTextures(1, texture);
   glBindTexture(GL_TEXTURE_2D, *texture);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB10_A2, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   //Bind texture to attachment
   glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachment, GL_TEXTURE_2D, *texture, 0);
   printf("DONE\n");
}

void
DeferredShadingEngine::initShaders()
{
   //GATHER SHADER
   gatherShader = new Shader();
   gatherShader->addVert(GATHER_VERT_SHADER);
   gatherShader->addFrag(GATHER_FRAG_SHADER);
   gatherShader->build();
   //DIRECT TARGET SHADER
   directTargetShader = new Shader();
   directTargetShader->addFrag(DIR_TARGET_FRAG_SHADER);
   directTargetShader->build();
   //DIRECTIONAL AND AMBIENT LIGHT SHADER
   dirAmbLightShader = new Shader();
   dirAmbLightShader->addFrag(DIR_AMB_LIGHT_FRAG_SHADER);
   dirAmbLightShader->build();
   //POINT LIGHT SHADER
   pointLightShader = new Shader();
   pointLightShader->addVert(POINT_LIGHT_VERT_SHADER);
   pointLightShader->addFrag(POINT_LIGHT_FRAG_SHADER);
   pointLightShader->build();
}

void
DeferredShadingEngine::preDrawScene()
{
   glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
   glPushAttrib(GL_VIEWPORT_BIT);
   glViewport(0,0,width, height);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //Must have this after switching to the FBO!
   gatherShader->attach();
}

void
DeferredShadingEngine::postDrawScene()
{
   gatherShader->detach();
   glPopAttrib();
   glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void
DeferredShadingEngine::drawFullScreenQuad(int screenWidth, int screenHeight)
{
   //Matrix setup
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glOrtho(0,screenWidth,0,screenHeight,-1,1);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();

   //Set OpenGL settings
   glColor3f(0,0,1);
   GLboolean depthTestEnabled;
   glGetBooleanv(GL_DEPTH_TEST, &depthTestEnabled);
   glDisable(GL_DEPTH_TEST);
   GLboolean lightingEnabled;
   glGetBooleanv(GL_LIGHTING, &lightingEnabled);
   glDisable(GL_LIGHTING);

   //Draw a Big Quad
   glBegin(GL_QUADS);
   glNormal3f( 0.0f, 0.0f, 1.0);
   glVertex3f(0, 0,  0);
   glVertex3f(screenWidth, 0,  0);
   glVertex3f(screenWidth, screenHeight,  0);
   glVertex3f(0, screenHeight,  0);
   glEnd();

   //Restore OpenGL settings
   if(depthTestEnabled)
      glEnable(GL_DEPTH_TEST);
   if(lightingEnabled)
      glEnable(GL_LIGHTING);

   //Matrix breakdown
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
}

void
DeferredShadingEngine::drawDirectionalAndAmbient(int screenWidth, int screenHeight)
{
   dirAmbLightShader->attach();
   //Set shader uniforms
   GLint loc = glGetUniformLocation(dirAmbLightShader->shaderProg, "normals"); //Get uniform sampler location
   glActiveTexture(GL_TEXTURE0);                    // Bind         to         unit
   glBindTexture(GL_TEXTURE_2D, textures[NORMALS]); //      normals    texture      0
   glUniform1i(loc, 0); //Pass 0 to denote texture unit 0
   loc = glGetUniformLocation(dirAmbLightShader->shaderProg, "colors"); //Get uniform sampler location
   glActiveTexture(GL_TEXTURE1);                   // Bind        to         unit
   glBindTexture(GL_TEXTURE_2D, textures[COLORS]); //      colors    texture      1
   glUniform1i(loc, 1); //Pass 1 to denote texture unit 1
   loc = glGetUniformLocation(dirAmbLightShader->shaderProg, "screenWidth");
   glUniform1i(loc, screenWidth);
   loc = glGetUniformLocation(dirAmbLightShader->shaderProg, "screenHeight");
   glUniform1i(loc, screenHeight);
   float lightDir[3] = {0.0, 0.0, 0.0};
   loc = glGetUniformLocation(dirAmbLightShader->shaderProg, "lightDirection");
   glUniform3fv(loc, 1, lightDir);
   float lightColor[3] = {0.5, 0.5, 0.5};
   loc = glGetUniformLocation(dirAmbLightShader->shaderProg, "diffuseLightColor");
   glUniform3fv(loc, 1, lightColor);
   float ambientColor[3] = {0.2, 0.2, 0.2};
   loc = glGetUniformLocation(dirAmbLightShader->shaderProg, "ambientLightColor");
   glUniform3fv(loc, 1, ambientColor);
   // draw
   drawFullScreenQuad(screenWidth, screenHeight);
   //breakdown
   dirAmbLightShader->detach();
}

void
DeferredShadingEngine::preDrawPointLights(int screenWidth, int screenHeight, int zNear, int zFar)
{
   pointLightShader->attach();
   //Set shader uniforms
   GLint loc = glGetUniformLocation(pointLightShader->shaderProg, "normals"); //Get uniform sampler location
   glActiveTexture(GL_TEXTURE0);                    // Bind         to         unit
   glBindTexture(GL_TEXTURE_2D, textures[NORMALS]); //      normals    texture      0
   glUniform1i(loc, 0); //Pass 0 to denote texture unit 0
   loc = glGetUniformLocation(pointLightShader->shaderProg, "depths"); //Get uniform sampler location
   glActiveTexture(GL_TEXTURE1);               // Bind        to         unit
   glBindTexture(GL_TEXTURE_2D, depthtexture); //      depths    texture      1
   glUniform1i(loc, 1); //Pass 1 to denote texture unit 1
   loc = glGetUniformLocation(pointLightShader->shaderProg, "colors"); //Get uniform sampler location
   glActiveTexture(GL_TEXTURE2);                   // Bind        to         unit
   glBindTexture(GL_TEXTURE_2D, textures[COLORS]); //      colors    texture      2
   glUniform1i(loc, 2); //Pass 2 to denote texture unit 2
   loc = glGetUniformLocation(pointLightShader->shaderProg, "screenWidth");
   glUniform1i(loc, screenWidth);
   loc = glGetUniformLocation(pointLightShader->shaderProg, "screenHeight");
   glUniform1i(loc, screenHeight);
   loc = glGetUniformLocation(pointLightShader->shaderProg, "near");
   glUniform1f(loc, zNear);
   loc = glGetUniformLocation(pointLightShader->shaderProg, "far");
   glUniform1f(loc, zFar);
   loc = glGetUniformLocation(pointLightShader->shaderProg, "radius");
   glUniform1f(loc, 2.0);
   loc = glGetUniformLocation(pointLightShader->shaderProg, "diffuseLightColor");
   glUniform3fv(loc, 1, lightColor);
   //setup required OpenGL state
   glFrontFace(GL_CW);
   glDisable(GL_DEPTH_TEST);
   glEnable(GL_BLEND);
   glBlendFunc(GL_ONE, GL_ONE);
}

void
DeferredShadingEngine::setPointLightColor(float r, float g, float b) {
   lightColor[0] = r;
   lightColor[1] = g;
   lightColor[2] = b;
   GLint loc = glGetUniformLocation(pointLightShader->shaderProg, "diffuseLightColor");
   glUniform3fv(loc, 1, lightColor);
}

void
DeferredShadingEngine::postDrawPointLights()
{
   //return the OpenGL state back to a "normal" state.
   //TODO: record what these settings were and restore them to their previous values so I don't clobber other people's stuff.
   glDisable(GL_BLEND);
   glEnable(GL_DEPTH_TEST);
   glFrontFace(GL_CCW);
   pointLightShader->detach();
}

void
DeferredShadingEngine::drawNormals(int screenWidth, int screenHeight)
{
   drawDirectTarget(textures[NORMALS], screenWidth, screenHeight);
}

void 
DeferredShadingEngine::drawColors(int screenWidth, int screenHeight)
{
   drawDirectTarget(textures[COLORS], screenWidth, screenHeight);
}

void
DeferredShadingEngine::drawDepth(int screenWidth, int screenHeight, float zNear, float zFar)
{
   drawDirectTarget(depthtexture, screenWidth, screenHeight, zNear, zFar);
}

void
DeferredShadingEngine::drawDirectTarget(GLuint bufferToDraw, int screenWidth, int screenHeight, float zNear, float zFar) {
   directTargetShader->attach();
   //Set shader uniforms
   GLint loc = glGetUniformLocation(directTargetShader->shaderProg, "bufferToDraw"); //Get uniform sampler location
   glActiveTexture(GL_TEXTURE0);               // Bind         to         unit
   glBindTexture(GL_TEXTURE_2D, bufferToDraw); //      texture    texture      0
   glUniform1i(loc, 0); //pass 0 to denote texture unit 0
   loc = glGetUniformLocation(directTargetShader->shaderProg, "screenWidth");
   glUniform1i(loc, screenWidth);
   loc = glGetUniformLocation(directTargetShader->shaderProg, "screenHeight");
   glUniform1i(loc, screenHeight);
   loc = glGetUniformLocation(directTargetShader->shaderProg, "near");
   glUniform1f(loc, zNear);
   loc = glGetUniformLocation(directTargetShader->shaderProg, "far");
   glUniform1f(loc, zFar);
   loc = glGetUniformLocation(directTargetShader->shaderProg, "doDepth");
   glUniform1i(loc, bufferToDraw==depthtexture);
   // draw
   drawFullScreenQuad(screenWidth, screenHeight);
   //breakdown
   directTargetShader->detach();
}
void
DeferredShadingEngine::checkExtensions()
{
   printf("***DeferredShadingEngine EXTENSION CHECK START***\n\n");

   //Check for OpenGL 2.0 for shaders and stuffs
   if (!glewIsSupported("GL_VERSION_2_0")) {
      printf("\tERROR: OpenGL 2.0 not supported.\n");
      printf("\t%s\n", (char*)glGetString( GL_VERSION ) );
      exit(-1);
   }
   printf("\tOpenGL 2.0 supported.\n");

   //Check for number of available buffers
   GLint maxbuffers;
   glGetIntegerv(GL_MAX_DRAW_BUFFERS, &maxbuffers);
   if(maxbuffers < NUM_MRTS) {
      printf("\tERROR: fewer than 4 render targets available.\n");
      exit(-1);
   }
   printf("\t%d render targets available.\n", maxbuffers);

   printf("\n***DeferredShadingEngine EXTENSION CHECK DONE!***\n");
}