/**
 * Library for Deferred Shading.
 * -----------------------------------
 * author: Ryan Schmitt
 * date: 2-20-11
 */

#ifndef __DEFERRED_SHADING_H__
#define __DEFERRED_SHADING_H__

#include <glew.h>
#include <glut.h>
#include "Shader.h"

class DeferredShadingEngine {
public:

   //---------------------------------------------------------------------------
   // CONSTRUCTOR
   //---------------------------------------------------------------------------

   /**
    * Construct a new DeferredShadingEngine.
    * @param width width of the buffers
    * @param height height of the buffers
    */
   DeferredShadingEngine(int width, int height);
   ~DeferredShadingEngine();

   //---------------------------------------------------------------------------
   // FUNCTIONS
   //---------------------------------------------------------------------------

   /**
    * Initialization function that must be called after all GL, GLUT, and GLEW
    * setup, but before use of other library functions.
    */
   void init();

   /**
    * Call to bind FBO, attach the gather shader, and perform other setup before draw calls.
    */
   void preDrawScene();

   /**
    * Call to unbind FBO, detach the gather shader, and perform other breakdown after draw calls.
    */
   void postDrawScene();

   /**
    * Draw the directional and ambient lighting.
    */
   void drawDirectionalAndAmbient(int screenWidth, int screenHeight);

   void preDrawPointLights(int screenWidth, int screenHeight, int zNear, int zFar);
   void setPointLightColor(float r, float g, float b);
   void postDrawPointLights();

   /**
    * Draws the normal buffer.
    */
   void drawNormals(int screenWidth, int screenHeight);

   /**
    * Draws the color buffer.
    */
   void drawColors(int screenWidth, int screenHeight);

   /**
    * Draws the depth buffer.
    */
   void drawDepth(int screenWidth, int screenHeight, float zNear, float zFar);

private:
   //---------------------------------------------------------------------------
   // UTILITY FUNCTIONS
   //---------------------------------------------------------------------------
   void checkExtensions();
   void checkBoundTexture();
   void initDepthTexture();
   void initRenderTargetTexture(const char* name, GLuint* texture, GLenum attachment);
   void initShaders();
   void drawFullScreenQuad(int screenWidth, int screenHeight);
   void drawDirectTarget(GLuint bufferToDraw, int screenWidth, int screenHeight, float zNear=0.0, float zFar=0.0);
   void drawLightGeometry();

   //---------------------------------------------------------------------------
   // VARIABLES
   //---------------------------------------------------------------------------
   float lightColor[3]; //TODO: remove this asap
   enum   MRT{NORMALS, COLORS, NUM_MRTS};
   GLuint textures[NUM_MRTS];
   GLuint fbo;
   GLuint  depthtexture;
   int  width, height; //buffer size
   Shader* dirAmbLightShader;
   Shader* pointLightShader;
   Shader* directTargetShader;
   Shader* gatherShader;
   static const char* GATHER_VERT_SHADER;
   static const char* GATHER_FRAG_SHADER;
   static const char* DIR_TARGET_FRAG_SHADER;
   static const char* DIR_AMB_LIGHT_FRAG_SHADER;
   static const char* POINT_LIGHT_VERT_SHADER;
   static const char* POINT_LIGHT_FRAG_SHADER;
};

#endif